# __main__.py

import sys
import argparse

import pandas as pd


def main():
    print("Project is on the way to update this feature...,\nThank you all for your support")
    return


if __name__ == '__main__':
    main()
